@@ -25,6 +25,7 @@
 
 #include "cmakeconfig.h"
 
+#include "BuildRevision.h"
 #include "HeadlessViewBackend.h"
 #include "WindowViewBackend.h"
 #include <memory>
@@ -135,11 +136,11 @@ static std::unique_ptr<WPEToolingBackends::ViewBackend> createViewBackend(uint32
     return std::make_unique<WPEToolingBackends::WindowViewBackend>(width, height);
 }
 
-typedef struct {
+struct FilterSaveData {
     GMainLoop* mainLoop { nullptr };
     WebKitUserContentFilter* filter { nullptr };
     GError* error { nullptr };
-} FilterSaveData;
+};
 
 static void filterSavedCallback(WebKitUserContentFilterStore *store, GAsyncResult *result, FilterSaveData *data)
 {
@@ -200,8 +201,8 @@ int main(int argc, char *argv[])
             webkit_get_major_version(),
             webkit_get_minor_version(),
             webkit_get_micro_version());
-        if (g_strcmp0(SVN_REVISION, "tarball"))
-            g_print(" (%s)", SVN_REVISION);
+        if (g_strcmp0(BUILD_REVISION, "tarball"))
+            g_print(" (%s)", BUILD_REVISION);
         g_print("\n");
         return 0;
     }
